import os
import functools
import binascii
import multihash
import multibase

from PyQt5.QtGui import QPixmap
from PyQt5.QtGui import QImage
from PyQt5.QtGui import QIcon

from PyQt5.QtCore import QStandardPaths
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QSize
from PyQt5.QtCore import QEvent
from PyQt5.QtCore import QObject
from PyQt5.QtCore import QDir
from PyQt5.QtCore import pyqtSignal

from PyQt5.QtWidgets import QStyle
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QInputDialog
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QVBoxLayout

from galacteek import ensure
from galacteek.ipfs.mimetype import mimeTypeDagUnknown
from galacteek.ipfs.mimetype import mimeTypeDagPb
from galacteek.ipfs.cidhelpers import getCID
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs import ipfsOpFn
from galacteek.ipfs import kilobytes


from .i18n import iIpfsQrCodes
from .i18n import iUnknown


def getIcon(iconName):
    app = QApplication.instance()
    if iconName in app._icons:
        return app._icons[iconName]

    if iconName.startswith(':/'):
        iconPath = iconName
    else:
        iconPath = ':/share/icons/{}'.format(iconName)

    icon = app._icons[iconName] = QIcon(QPixmap(iconPath))
    return icon


def getMimeIcon(mType):
    app = QApplication.instance()
    return app.mimeTypeIcons.get(mType)


def preloadMimeIcons():
    icons = {}
    mIconsDir = QDir(':/share/icons/mimetypes')

    if not mIconsDir.exists():
        return icons

    entries = mIconsDir.entryList()

    for entry in entries:
        mType = entry.replace('.png', '').replace('-', '/', 1)
        icons[mType] = getIcon('mimetypes/{}'.format(entry))

    return icons


async def getIconFromIpfs(ipfsop, ipfsPath, scaleWidth=None,
                          sizeMax=kilobytes(256), timeout=10):
    """
    We cache the icons that we got out of the IPFS repo by their path
    Max icons cached is set by 'ipfsIconsCacheMax' in the app object
    """
    app = QApplication.instance()
    iconsCache = app.ipfsIconsCache

    if ipfsPath in iconsCache:
        # Already cached
        return iconsCache[ipfsPath]

    if len(iconsCache) >= app.ipfsIconsCacheMax:
        # FIFO 1/8
        for icount in range(0, int(app.ipfsIconsCacheMax / 8)):
            out = list(iconsCache.keys())[0]
            del iconsCache[out]

    try:
        mimeType, stat = await app.rscAnalyzer(ipfsPath)
        if not mimeType or not stat:
            return None

        statInfo = StatInfo(stat)

        if not statInfo.valid or statInfo.dataLargerThan(sizeMax):
            return None
        elif not mimeType.isImage:
            return None

        imgData = await ipfsop.waitFor(
            ipfsop.client.cat(ipfsPath), timeout
        )

        if not imgData:
            return None

        icon = getIconFromImageData(imgData, scaleWidth=scaleWidth)
        if icon:
            iconsCache[ipfsPath] = icon
            return icon
    except BaseException:
        return None


async def getFavIconFromDir(ipfsop, ipfsPath, timeout=10):
    """
    If a favicon.ico file exists inside the given directory,
    return it in the form of a QIcon
    """

    faviconPath = ipfsPath.child('favicon.ico')
    try:
        stat = await ipfsop.objStat(str(faviconPath))

        if stat:
            # favicon exists
            return await getIconFromIpfs(ipfsop, str(faviconPath))
    except BaseException:
        return None


def getIconFromImageData(imgData, scaleWidth=None):
    try:
        img = QImage()
        img.loadFromData(imgData)

        if isinstance(scaleWidth, int):
            img = img.scaledToWidth(scaleWidth)

        return QIcon(QPixmap.fromImage(img))
    except BaseException:
        return None


@ipfsOpFn
async def getImageFromIpfs(ipfsop, ipfsPath, timeout=10):
    try:
        imgData = await ipfsop.waitFor(
            ipfsop.client.cat(str(ipfsPath)), timeout
        )

        if not imgData:
            return None

        image = QImage()
        image.loadFromData(imgData)
        return image
    except BaseException:
        return None


def getIconIpfsIce():
    return getIcon('ipfs-logo-128-ice.png')


def getIconIpfsWhite():
    return getIcon('ipfs-logo-128-white.png')


def getIconIpfs64():
    return getIcon('ipfs-cube-64.png')


def getIconClipboard():
    return getIcon('clipboard.png')


def getIconFromMimeType(mimeType, defaultIcon=None):
    if not mimeType.valid:
        return getMimeIcon('unknown')

    fIcon = getMimeIcon(mimeType.type)
    if fIcon:
        return fIcon

    mIcon = None
    if mimeType.isDir:
        mIcon = 'inode/directory'
    elif mimeType.isHtml:
        mIcon = 'text/html'
    elif mimeType.isText:
        mIcon = 'text/plain'
    elif mimeType.isImage:
        mIcon = 'image/x-generic'
    elif mimeType.isVideo:
        mIcon = 'video/x-generic'
    elif mimeType.isAudio:
        mIcon = 'audio/x-generic'
    elif mimeType in [mimeTypeDagUnknown, mimeTypeDagPb]:
        mIcon = mimeType.type

    icon = getMimeIcon(mIcon)
    return icon if icon else getMimeIcon(
        defaultIcon if defaultIcon else 'unknown')


def getHomePath():
    pList = QStandardPaths.standardLocations(QStandardPaths.HomeLocation)
    return pList[0] if len(pList) > 0 else os.getenv('HOME')


def messageBox(message, title=None):
    from .clips import RotatingCubeClipSimple
    from .widgets import AnimatedButton

    app = QApplication.instance()
    msgBox = QDialog()

    largeSize = app.style().pixelMetric(QStyle.PM_LargeIconSize)

    button = AnimatedButton(
        RotatingCubeClipSimple(), ignoreFrameEvery=2, parent=msgBox)
    button.clicked.connect(functools.partial(msgBox.done, 1))
    button.setIconSize(QSize(largeSize, largeSize))
    button.clip.setSpeed(100)
    button.setText('OK')
    button.startClip()

    label = QLabel(message)
    label.setAlignment(Qt.AlignCenter)
    layout = QVBoxLayout()
    layout.addWidget(label)
    layout.addWidget(button)
    msgBox.setLayout(layout)
    msgBox.setMinimumSize(
        app.desktopGeometry.width() / 3,
        128
    )

    msgBox.setWindowTitle(title if title else 'galacteek: Message')
    msgBox.show()
    return msgBox.exec_()


def questionBox(title, text, parent=None):
    box = QMessageBox.question(parent, title, text)
    return box == QMessageBox.Yes


def directorySelect(caption=''):
    return QFileDialog.getExistingDirectory(
        None, caption, getHomePath(), QFileDialog.ShowDirsOnly)


def filesSelect(filter='(*.*)'):
    result = QFileDialog.getOpenFileNames(None,
                                          '', getHomePath(), filter)
    if result:
        return result[0]


def filesSelectImages():
    return filesSelect(
        filter='Images (*.xpm *.jpg *.jpeg *.png *.svg)')


def saveFileSelect(filter='(*.*)'):
    result = QFileDialog.getSaveFileName(None,
                                         '', getHomePath(), filter)
    if result:
        return result[0]


def disconnectSig(sig, target):
    try:
        sig.disconnect(target)
    except Exception:
        pass


def runDialog(cls, *args, **kw):
    title = kw.pop('title', None)
    accepted = kw.pop('accepted', None)
    dlgW = cls(*args, **kw)

    def onAccept():
        if accepted:
            accepted(dlgW)
    if title:
        dlgW.setWindowTitle(title)
    if accepted:
        dlgW.accepted.connect(functools.partial(accepted, dlgW))

    dlgW.show()
    dlgW.exec_()
    return dlgW


def inputText(title='', label='', parent=None):
    text, ok = QInputDialog.getText(parent, title, label)
    if ok:
        return text


def inputTextLong(title='', label='', text='', parent=None):
    text, ok = QInputDialog.getText(
        parent, title, label, QLineEdit.Normal,
        text, Qt.Dialog, Qt.ImhNone)
    if ok:
        return text


class IPFSTreeKeyFilter(QObject):
    deletePressed = pyqtSignal()
    copyHashPressed = pyqtSignal()
    copyPathPressed = pyqtSignal()
    explorePressed = pyqtSignal()
    returnPressed = pyqtSignal()
    backspacePressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if key == Qt.Key_Return:
                self.returnPressed.emit()
                return True
            if key == Qt.Key_Backspace:
                self.backspacePressed.emit()
                return True
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_C or key == Qt.Key_Y:
                    self.copyPathPressed.emit()
                    return True
                if key == Qt.Key_A:
                    self.copyHashPressed.emit()
                    return True
                if key == Qt.Key_X:
                    self.explorePressed.emit()
                    return True
            if event.key() == Qt.Key_Delete:
                self.deletePressed.emit()
                return True
        return False


class BasicKeyFilter(QObject):
    deletePressed = pyqtSignal()
    copyPressed = pyqtSignal()
    returnPressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if key == Qt.Key_Return:
                self.returnPressed.emit()
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_C:
                    self.copyPressed.emit()

            if event.key() == Qt.Key_Delete:
                self.deletePressed.emit()
            return True
        return False


def clipboardSupportsSelection():
    return QApplication.clipboard().supportsSelection()


def sizeFormat(size):
    if size > (1024 * 1024 * 1024):
        return '{0:.2f} Gb'.format(size / (1024 * 1024 * 1024))
    if size > (1024 * 1024):
        return '{0:.2f} Mb'.format(size / (1024 * 1024))
    if size > 1024:
        return '{0:.2f} kb'.format(size / 1024)
    if size == 0:
        return '0'
    return '{0} bytes'.format(size)


def qrCodesMenuBuilder(urls, resourceOpener, parent=None):
    if isinstance(urls, list):
        icon = getIcon('ipfs-qrcode.png')
        menu = QMenu(iIpfsQrCodes(), parent)
        menu.setIcon(icon)

        for url in urls:
            menu.addAction(
                icon, str(url), functools.partial(
                    ensure, resourceOpener.open(url)))
        return menu


def cidInfosMarkup(cidString):
    try:
        cid = getCID(cidString)
        mhash = multihash.decode(cid.multihash)
        baseEncoding = multibase.multibase.get_codec(cidString)
    except:
        return '<p>Invalid CID</p>'

    return '''
    <p>CID: <b>{cids}</b></p>
    <p>CID version {cidv}</p>
    <p>Multibase encoding: <b>{mbase}</b></p>

    <p>Codec: <b>{codec}</b></p>
    <p>Multihash function: <b>{mhashfunc}</b></p>
    <p>Multihash function code: {mhashfuncvalue}
        (<b>{mhashfuncvaluehex}</b>)</p>
    <p>Multihash digest: <b>{mhashdigest}</b></p>
    <p>Multihash: <b>{mhashascii}</b></p>
    '''.format(cids=cidString,
               cidv=cid.version,
               mbase=baseEncoding.encoding if baseEncoding else iUnknown(),
               codec=cid.codec,
               mhashfunc=mhash.func.name,
               mhashfuncvalue=mhash.func.value,
               mhashfuncvaluehex=hex(mhash.func.value),
               mhashdigest=binascii.b2a_hex(mhash.digest).decode('ascii'),
               mhashascii=binascii.b2a_hex(cid.multihash).decode('ascii'))



import os
import os.path
import re
import sys
import codecs
import subprocess
import glob
import shutil
import json

from setuptools import setup
from setuptools import Command
from distutils.command.build import build

PY_VER = sys.version_info

if PY_VER >= (3, 6):
    pass
else:
    print('You need python3.6 or newer')
    print('Your python version is {0}'.format(PY_VER))
    raise RuntimeError('Invalid python version')

with codecs.open(os.path.join(os.path.abspath(os.path.dirname(
        __file__)), 'galacteek', '__init__.py'), 'r', 'latin1') as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


def run(*args):
    p = subprocess.Popen(*args, stdout=subprocess.PIPE)
    stdout, err = p.communicate()
    return stdout


class build_docs(Command):
    user_options = [
        ("all=", None, "Build all docs"),
    ]

    def initialize_options(self):
        self.all = None

    def finalize_options(self):
        pass

    def run(self):
        from sphinx import build_main
        args = [sys.argv[0], '-b', 'html',
                'galacteek/docs/manual/en',
                'galacteek/docs/manual/en/html']

        if self.all:
            args.append('-a')

        build_main(args)


class build_contracts(Command):
    user_options = [
        ("deploy=", None, "Deploy given contracts"),
        ("contracts=", None, "Contracts list to build"),
        ("rpcurl=", None, "Ethereum RPC url"),
    ]

    def initialize_options(self):
        self.deploy = None
        self.contracts = None
        self.rpcurl = 'http://127.0.0.1:7545'

    def finalize_options(self):
        pass

    def run(self):
        from web3 import Web3
        from galacteek.smartcontracts import listContracts
        from galacteek.smartcontracts import solCompileFile
        from galacteek.smartcontracts import vyperCompileFile
        from galacteek.dweb.ethereum.contract import contractDeploy

        usrcontracts = [c for c in self.contracts.split(',')] if \
            self.contracts else ['*']
        cdeploy = [c for c in self.deploy.split(',')] if \
            self.deploy else []

        w3 = Web3(Web3.HTTPProvider(self.rpcurl))
        w3.eth.defaultAccount = w3.eth.accounts[0]

        for contract in listContracts():
            print('>', contract, contract.dir)

            if usrcontracts != ['*'] and contract not in usrcontracts:
                continue

            ifacePath = os.path.join(contract.dir, 'interface.json')

            if contract.type == 'vyper':
                iface = vyperCompileFile(contract.sourcePath)
                if not iface:
                    print('Error compiling vyper contract')
                    continue
            elif contract.type == 'solidity':
                compiled = solCompileFile(contract.sourcePath)
                if not compiled:
                    print('Error compiling solidity contract')
                    continue
                contractId, iface = compiled.popitem()
            else:
                continue

            try:
                with open(ifacePath, 'w+t') as ifacefd:
                    ifacefd.write(json.dumps(iface, indent=4))
            except Exception as err:
                print(str(err))
            else:
                print(contract.name, 'compiled')
                if contract.name in cdeploy:
                    addr = contractDeploy(w3, iface)
                    print(contract.name, 'deployed at', addr)


class build_ui(Command):
    user_options = [
        ("uiforms=", None, "UI forms list to build, separated by ','"),
    ]

    def initialize_options(self):
        self.uiforms = None

        # Forms where we don't want to have automatic slots
        # connection with connectSlotsByName()
        self.uiforms_noSlotConnect = [
            'galacteek/ui/browsertab.ui',
            'galacteek/ui/dagview.ui',
            'galacteek/ui/files.ui',
            'galacteek/ui/qschemecreatemapping.ui'
        ]

    def finalize_options(self):
        pass

    def filterUic(self, uifile, uicpath):
        if uifile in self.uiforms_noSlotConnect:
            print('* {ui}: Removing automatic slots connection'.format(
                ui=uifile))

            with open(uicpath, 'rt') as fd:
                code = fd.read()

            nCode = re.sub(
                r'^\s*QtCore.QMetaObject.connectSlotsByName.*\n$', '',
                code,
                flags=re.MULTILINE
            )

            with open(uicpath, 'wt') as fd:
                print('* {ui}: Rewriting {path}'.format(
                    ui=uifile, path=uicpath))
                fd.write(nCode)

    def run(self):
        uifiles = []
        uidir = 'galacteek/ui'

        if self.uiforms:
            uifiles = [os.path.join(uidir, '{0}.ui'.format(form)) for form
                       in self.uiforms.split(',')]
        else:
            uifiles = glob.iglob('{}/*.ui'.format(uidir))

        for uifile in uifiles:
            print('* Building UI form:', uifile)
            base = os.path.basename(uifile).replace('.ui', '')
            out = 'ui_{}.py'.format(base)
            fp_out = os.path.join(uidir, out)

            run(['pyuic5', '--from-imports',
                 uifile,
                 '-o', fp_out])

            self.filterUic(uifile, fp_out)

        run(['pylupdate5', '-verbose', 'galacteek.pro'])

        trdir = './share/translations'
        lrelease = shutil.which('lrelease-qt5')

        if not lrelease:
            lrelease = shutil.which('lrelease')

        for lang in ['en', 'fr']:
            if lrelease:
                run([lrelease,
                     os.path.join(trdir, 'galacteek_{}.ts'.format(lang)),
                     '-qm',
                     os.path.join(trdir, 'galacteek_{}.qm'.format(lang))])
            else:
                print('lrelease was not found, cannot build translation files')

        run(['pyrcc5', os.path.join(uidir, 'galacteek.qrc'), '-o',
             os.path.join(uidir, 'galacteek_rc.py')])


class _build(build):
    sub_commands = [('build_ui', None)] + build.sub_commands


with open('README.rst', 'r') as fh:
    long_description = fh.read()

deps_links = []
install_reqs = []
with open('requirements.txt') as f:
    lines = f.read().splitlines()
    for line in lines:
        if line.startswith('-e'):
            link = line.split().pop()
            deps_links.append(link)
        else:
            install_reqs.append(line)

setup(
    name='galacteek',
    version=version,
    license='GPL3',
    author='David Ferlier',
    author_email='galacteek@protonmail.com',
    url='https://github.com/eversum/galacteek',
    description='Browser for the distributed web',
    long_description=long_description,
    include_package_data=True,
    cmdclass={
        'build': _build,
        'build_ui': build_ui,
        'build_docs': build_docs,
        'build_contracts': build_contracts
    },
    packages=[
        'galacteek',
        'galacteek.docs',
        'galacteek.docs.manual',
        'galacteek.core',
        'galacteek.core.models',
        'galacteek.crypto',
        'galacteek.ipfs',
        'galacteek.ipfs.pubsub',
        'galacteek.ipfs.pb',
        'galacteek.ipfs.p2pservices',
        'galacteek.ipdapps',
        'galacteek.hashmarks',
        'galacteek.hashmarks.default',
        'galacteek.templates',
        'galacteek.dweb',
        'galacteek.dweb.ethereum',
        'galacteek.smartcontracts',
        'galacteek.ui',
        'galacteek.ui.orbital'
    ],
    install_requires=install_reqs,
    dependency_links=deps_links,
    package_data={
        'galacteek': [
            'docs/manual/en/html/*.html',
            'docs/manual/en/html/_images/*',
            'docs/manual/en/html/_static/*',
            'templates/*.html',
            'templates/layouts/*',
            'templates/usersite/*.html',
            'templates/usersite/assets/*',
            'templates/usersite/assets/css/*',
            'templates/usersite/macros/*',
            'templates/imggallery/*.html',
            'hashmarks/default/*.json'
        ]
    },
    entry_points={
        'gui_scripts': [
            'galacteek = galacteek.guientrypoint:start'
        ]
    },
    extras_require={
        'docs': [
            'sphinx>=1.7.0'
        ]
    },
    classifiers=[
        'Environment :: X11 Applications :: Qt',
        'Framework :: AsyncIO',
        'Topic :: Desktop Environment :: File Managers',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: System :: Filesystems',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords=[
        'asyncio',
        'aiohttp',
        'ipfs'
    ]
)



import pkg_resources
import importlib
import os.path
import os
import jinja2

from PyQt5.QtCore import QTimer

from galacteek import log
from galacteek.core.objects import GObject
from galacteek.core.objects import pyqtSignal
from galacteek.core import runningApp
from galacteek.dweb import render
from galacteek.ipfs import ipfsOp
from galacteek.core.fswatcher import FileWatcher

dappsPkg = __name__


def availableDapps():
    _dapps = {}
    try:
        listing = pkg_resources.resource_listdir(dappsPkg, '')
        for fn in listing:
            path = pkg_resources.resource_filename(dappsPkg, fn)
            if not fn.startswith('_') and os.path.isdir(path):
                _dapps[fn] = path
        return _dapps
    except Exception:
        pass


def dappGetModule(dappName):
    modname = '{0}.{1}'.format(dappsPkg, dappName)

    try:
        module = importlib.import_module(modname)
    except Exception as err:
        print(str(err))
        return None
    else:
        return module


def dappsRegisterSchemes():
    # Called early (register dapps schemes before app creation)
    dapps = availableDapps()
    for dappName in dapps:
        mod = dappGetModule(dappName)
        if mod:
            print(dappName, mod, 'reg schemes')
            mod.declareUrlSchemes()


class IPLiveDapp(GObject):
    """
    Inter-Planetary Dapp
    """

    pkgContentsChanged = pyqtSignal(str)

    def __init__(self, ethCtrl, name, pkgPath,
                 offline=False, **kwargs):
        parent = kwargs.pop('parent', None)
        planet = kwargs.pop('planet', 'earth')
        initmod = kwargs.pop('initModule', None)

        super(IPLiveDapp, self).__init__(parent)

        self.app = runningApp()

        self._ethController = ethCtrl
        self._name = name
        self._pkgPath = pkgPath
        self._templatesPath = os.path.join(self._pkgPath, 'templates')
        self._assetsPath = os.path.join(self._pkgPath, 'assets')
        self._assetsCid = None
        self._args = kwargs
        self._planet = planet
        self._offline = offline

        self._ctx_params = {}

        self.jEnv = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self._templatesPath),
            enable_async=True
        )

        self._moduleInit = initmod
        self._watcher = FileWatcher()
        self._watchT = QTimer(self)
        self._watchT.timeout.connect(
            lambda: self.pkgContentsChanged.emit(self.name))

        self.watch(self._pkgPath)
        self._watcher.pathChanged.connect(self.onFileWatchChange)

    @property
    def name(self):
        return self._name

    @property
    def offline(self):
        # DAGs and assets in 'offline' mode (dev)
        return self._offline

    @property
    def planet(self):
        return self._planet

    @property
    def eth(self):
        return self._ethController

    @property
    def web3(self):
        return self.eth.web3

    @property
    def args(self):
        return self._args

    @property
    def assetsCid(self):
        return self._assetsCid

    async def registerSchemeHandlers(self):
        pass

    def watch(self, _dir):
        for root, dirs, files in os.walk(_dir):
            for d in dirs:
                if d.startswith('__'):
                    continue
                fp = os.path.join(root, d)
                self._watcher.watch(fp)
        self._watcher.watch(_dir)

    def compileTemplates(self):
        try:
            self.jEnv.compile_templates(self._templatesCompiledPath, zip=None)
        except Exception as err:
            print('Templates recompile error', str(err))

    @ipfsOp
    async def init(self, ipfsop):
        self.compileTemplates()
        await self.importAssets()

    def stop(self):
        pass

    @ipfsOp
    async def importAssets(self, ipfsop):
        if os.path.isdir(self._assetsPath):
            self._assetsCid = await ipfsop.addPath(
                self._assetsPath, recursive=True,
                offline=self.offline
            )

    async def renderTemplate(self, template, **kw):
        coro = render.ipfsRender
        kw.update(self._ctx_params)
        kw['offline'] = self.offline

        return await coro(self.jEnv,
                          template,
                          **kw
                          )

    async def renderTemplateWLink(self, dag, tmpl, **kw):
        # Render and link in the dag
        result = await self.renderTemplate(tmpl, **kw)
        return dag.mkLink(result)

    def onFileWatchChange(self, path):
        self._watchT.stop()
        self._watchT.start(3000)

    def reimportModule(self):
        self._watchT.stop()
        if self._moduleInit:
            try:
                importlib.reload(self._moduleInit)
            except ImportError:
                log.debug('Error reimporting module {}'.format(
                    self._moduleInit))

    async def dappReload(self):
        self._watchT.stop()
        log.debug('Dapp reload')

        await self.importAssets()
        try:
            self.jEnv.compile_templates(self._templatesCompiledPath)
        except Exception as err:
            print('Templates recompile error', str(err))



import os.path
import pkg_resources
import importlib
import json

from galacteek import log


pkg = __name__


def vyperCompileFile(file_path, formats=['abi', 'bytecode']):
    from vyper import compile_code
    try:
        with open(file_path, 'r') as f:
            source = f.read()

        return compile_code(source, formats)
    except Exception as err:
        log.debug(str(err))


def solCompileFile(file_path):
    from solcx import compile_source
    try:
        with open(file_path, 'r') as f:
            source = f.read()

        return compile_source(source)
    except Exception as err:
        log.debug(str(err))


class LocalContract:
    def __init__(self, name, rootDir, cPath, modImp, ctype='solidity'):
        self.name = name
        self.rootDir = rootDir
        self.sourcePath = cPath
        self._deployedAddress = None
        self._type = ctype
        self.module = modImp

    def interface(self):
        fp = os.path.join(self.dir, 'interface.json')
        if os.path.exists(fp):
            with open(fp, 'rt') as fd:
                return json.load(fd)

    def compile(self):
        if self.type == 'solidity':
            compiled = solCompileFile(self.sourcePath)
            _id, interface = compiled.popitem()
            return interface
        elif self.type == 'vyper':
            compiled = vyperCompileFile(self.sourcePath)
            return compiled

    def web3Contract(self, w3):
        iface = self.interface()

        if 'bytecode' in iface:
            binary = iface['bytecode']
        elif 'bin' in iface:
            binary = iface['bin']
        else:
            raise ValueError('Invalid interface')

        return w3.eth.contract(
            bytecode=binary,
            abi=iface.get('abi')
        )

    @property
    def type(self):
        return self._type

    @property
    def dir(self):
        return self.rootDir

    @property
    def address(self):
        return self._deployedAddress

    @address.setter
    def address(self, addr):
        self._deployedAddress = addr

    def __repr__(self):
        return 'Local contract: {}'.format(self.sourcePath)


def listContracts():
    try:
        listing = pkg_resources.resource_listdir(pkg, '')
        for fn in listing:
            contract = getContractByName(fn)
            if contract:
                yield contract
    except Exception:
        pass


def getContractByName(name):
    path = pkg_resources.resource_filename(pkg, name)
    mainSol = os.path.join(path, 'contract.sol')
    vyperPath = os.path.join(path, 'contract.vy')

    if not os.path.isdir(path):
        return None

    modname = '{0}.{1}'.format(pkg, name)

    try:
        module = importlib.import_module(modname)
    except Exception as err:
        print(str(err))
        return None

    if os.path.isfile(mainSol):
        return LocalContract(name, path, mainSol, module, ctype='solidity')
    elif os.path.isfile(vyperPath):
        return LocalContract(name, path, vyperPath, module, ctype='vyper')



from PyQt5.QtWidgets import QApplication
from PyQt5.QtWebEngineWidgets import QWebEngineProfile
from PyQt5.QtWebEngineWidgets import QWebEngineSettings
from PyQt5.QtWebEngineCore import QWebEngineUrlRequestInterceptor

from galacteek import log

from galacteek.dweb.webscripts import ethereumClientScripts

from galacteek.core.schemes import SCHEME_DWEB
from galacteek.core.schemes import SCHEME_ENS
from galacteek.core.schemes import SCHEME_ENSR
from galacteek.core.schemes import SCHEME_FS
from galacteek.core.schemes import SCHEME_IPFS
from galacteek.core.schemes import SCHEME_IPNS
from galacteek.core.schemes import SCHEME_Q


WP_NAME_MINIMAL = 'minimal'
WP_NAME_IPFS = 'ipfs'
WP_NAME_WEB3 = 'web3'


class IPFSRequestInterceptor(QWebEngineUrlRequestInterceptor):
    def interceptRequest(self, info):
        url = info.requestUrl()

        if url.isValid():
            log.debug('IPFS interceptor for URL: {}'.format(
                url.toString()))


class BaseProfile(QWebEngineProfile):
    def __init__(self, storageName='base', parent=None):
        super(BaseProfile, self).__init__(storageName, parent)

        self.app = QApplication.instance()
        self.webScripts = self.scripts()
        self.webSettings = self.settings()
        self.profileName = storageName
        self.installIpfsSchemeHandlers()
        self.installScripts()

        self.downloadRequested.connect(
            self.app.downloadsManager.onDownloadRequested)

    def installHandler(self, scheme, handler):
        sch = scheme if isinstance(scheme, bytes) else scheme.encode()
        self.installUrlSchemeHandler(sch, handler)

    def installScripts(self):
        pass

    def installIpfsSchemeHandlers(self):
        # XXX Remove fs: soon
        for scheme in [SCHEME_DWEB, SCHEME_FS]:
            self.installHandler(scheme, self.app.ipfsSchemeHandler)

        for scheme in [SCHEME_IPFS, SCHEME_IPNS]:
            self.installHandler(scheme, self.app.nativeIpfsSchemeHandler)

        self.installHandler(SCHEME_ENS, self.app.ensProxySchemeHandler)
        self.installHandler(SCHEME_ENSR, self.app.ensSchemeHandler)
        self.installHandler(SCHEME_Q, self.app.qSchemeHandler)


class MinimalProfile(BaseProfile):
    def __init__(self, storageName=WP_NAME_MINIMAL, parent=None):
        super(MinimalProfile, self).__init__(storageName, parent)


class IPFSProfile(BaseProfile):
    """
    IPFS web profile
    """

    def __init__(self, storageName=WP_NAME_IPFS, parent=None):
        super(IPFSProfile, self).__init__(storageName, parent)

        self.setSettings()

    def setSettings(self):
        self.webSettings.setAttribute(QWebEngineSettings.PluginsEnabled,
                                      True)
        self.webSettings.setAttribute(QWebEngineSettings.LocalStorageEnabled,
                                      True)

    def installScripts(self):
        exSc = self.webScripts.findScript('ipfs-http-client')
        if self.app.settingsMgr.jsIpfsApi is True and exSc.isNull():
            for script in self.app.scriptsIpfs:
                self.webScripts.insert(script)

    def installIpfsSchemeHandlersOld(self):
        # XXX Remove fs: soon
        for scheme in [SCHEME_DWEB, SCHEME_FS]:
            self.installUrlSchemeHandler(
                scheme.encode(), self.app.ipfsSchemeHandler)

        for scheme in [SCHEME_IPFS, SCHEME_IPNS]:
            self.installUrlSchemeHandler(
                scheme.encode(), self.app.nativeIpfsSchemeHandler)

        self.installUrlSchemeHandler(
            SCHEME_ENS.encode(), self.app.ensSchemeHandler)


class Web3Profile(IPFSProfile):
    """
    Web3 profile. Derives from the IPFS profile and
    adds an injection script to provide window.web3
    """

    def __init__(self, storageName=WP_NAME_WEB3, parent=None):
        super(Web3Profile, self).__init__(storageName, parent)

        self.installWeb3Scripts()

    def installWeb3Scripts(self):
        if self.app.settingsMgr.ethereumEnabled:
            ethereumScripts = ethereumClientScripts(
                self.app.getEthParams())
            if ethereumScripts:
                [self.webScripts.insert(script) for script in
                 ethereumScripts]
